/*
 * Copyright Microsoft Corp. 2023
 *
 * ch_interface.c: methods to connect guest interfaces to appropriate host
 * backends
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#pragma once


#include "ch_conf.h"
#include "virconftypes.h"


int
virCHConnetNetworkInterfaces(virCHDriver *driver,
                             virDomainDef *vmdef,
                             virDomainNetDef *netdef,
                             int *tapfds,
                             int **nicindexes,
                             size_t *nnicindexes);
